"""systemd logind mock template

This creates the expected methods and properties of the main
org.freedesktop.login1.Manager object. You can specify D-Bus property values
like "CanSuspend" or the return value of Inhibit() in "parameters".
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2013 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import os

import dbus
from gi.repository import GLib

from dbusmock import MOCK_IFACE, mockobject

BUS_NAME = "org.freedesktop.login1"
MAIN_OBJ = "/org/freedesktop/login1"
MAIN_IFACE = "org.freedesktop.login1.Manager"
SYSTEM_BUS = True


def load(mock, parameters):
    mock.AddMethods(
        MAIN_IFACE,
        [
            ("PowerOff", "b", "", ""),
            ("Reboot", "b", "", ""),
            ("Suspend", "b", "", ""),
            ("Hibernate", "b", "", ""),
            ("HybridSleep", "b", "", ""),
            ("SuspendThenHibernate", "b", "", ""),
            ("CanPowerOff", "", "s", f'ret = "{parameters.get("CanPowerOff", "yes")}"'),
            ("CanReboot", "", "s", f'ret = "{parameters.get("CanReboot", "yes")}"'),
            ("CanSuspend", "", "s", f'ret = "{parameters.get("CanSuspend", "yes")}"'),
            ("CanHibernate", "", "s", f'ret = "{parameters.get("CanHibernate", "yes")}"'),
            ("CanHybridSleep", "", "s", f'ret = "{parameters.get("CanHybridSleep", "yes")}"'),
            ("CanSuspendThenHibernate", "", "s", f'ret = "{parameters.get("CanSuspendThenHibernate", "yes")}"'),
            ("GetSession", "s", "o", 'ret = "/org/freedesktop/login1/session/" + args[0]'),
            ("ActivateSession", "s", "", ""),
            ("ActivateSessionOnSeat", "ss", "", ""),
            ("KillSession", "sss", "", ""),
            ("LockSession", "s", "", ""),
            ("LockSessions", "", "", ""),
            ("ReleaseSession", "s", "", ""),
            ("TerminateSession", "s", "", ""),
            ("UnlockSession", "s", "", ""),
            ("UnlockSessions", "", "", ""),
            ("GetSeat", "s", "o", 'ret = "/org/freedesktop/login1/seat/" + args[0]'),
            ("ListSeats", "", "a(so)", 'ret = [(k.split("/")[-1], k) for k in objects.keys() if "/seat/" in k]'),
            ("TerminateSeat", "s", "", ""),
            ("GetUser", "u", "o", 'ret = "/org/freedesktop/login1/user/" + args[0]'),
            ("KillUser", "us", "", ""),
            ("TerminateUser", "u", "", ""),
        ],
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "IdleHint": parameters.get("IdleHint", False),
                "IdleAction": parameters.get("IdleAction", "ignore"),
                "IdleSinceHint": dbus.UInt64(parameters.get("IdleSinceHint", 0)),
                "IdleSinceHintMonotonic": dbus.UInt64(parameters.get("IdleSinceHintMonotonic", 0)),
                "IdleActionUSec": dbus.UInt64(parameters.get("IdleActionUSec", 1)),
                "PreparingForShutdown": parameters.get("PreparingForShutdown", False),
                "PreparingForSleep": parameters.get("PreparingForSleep", False),
            },
            signature="sv",
        ),
    )


#
# logind methods which are too big for squeezing into AddMethod()
#


@dbus.service.method(MAIN_IFACE, in_signature="", out_signature="a(uso)")
def ListUsers(_):
    users = []
    for k, obj in mockobject.objects.items():
        if "/user/" in k:
            uid = dbus.UInt32(int(k.split("/")[-1]))
            users.append((uid, obj.Get("org.freedesktop.login1.User", "Name"), k))
    return users


@dbus.service.method(MAIN_IFACE, in_signature="", out_signature="a(susso)")
def ListSessions(_):
    sessions = []
    for k, obj in mockobject.objects.items():
        if "/session/" in k:
            session_id = k.split("/")[-1]
            uid = obj.Get("org.freedesktop.login1.Session", "User")[0]
            username = obj.Get("org.freedesktop.login1.Session", "Name")
            seat = obj.Get("org.freedesktop.login1.Session", "Seat")[0]
            sessions.append((session_id, uid, username, seat, k))
    return sessions


@dbus.service.method(MAIN_IFACE, in_signature="ssss", out_signature="h")
def Inhibit(_, what, who, why, mode):
    if not hasattr(mockobject, "inhibitors"):
        mockobject.inhibitors = []

    fd_r, fd_w = os.pipe()

    inhibitor = (what, who, why, mode, 1000, 123456)
    mockobject.inhibitors.append(inhibitor)

    def inhibitor_dropped(fd, cond):
        # pylint: disable=unused-argument
        os.close(fd)
        mockobject.inhibitors.remove(inhibitor)
        return False

    GLib.unix_fd_add_full(GLib.PRIORITY_HIGH, fd_r, GLib.IO_HUP, inhibitor_dropped)
    GLib.idle_add(os.close, fd_w)

    return fd_w


@dbus.service.method(MAIN_IFACE, in_signature="", out_signature="a(ssssuu)")
def ListInhibitors(_):
    if not hasattr(mockobject, "inhibitors"):
        mockobject.inhibitors = []

    return mockobject.inhibitors


#
# Convenience methods on the mock
#


@dbus.service.method(MOCK_IFACE, in_signature="s", out_signature="s")
def AddSeat(self, seat):
    """Convenience method to add a seat.

    Return the object path of the new seat.
    """
    seat_path = "/org/freedesktop/login1/seat/" + seat
    if seat_path in mockobject.objects:
        raise dbus.exceptions.DBusException(f"Seat {seat} already exists", name=MOCK_IFACE + ".SeatExists")

    self.AddObject(
        seat_path,
        "org.freedesktop.login1.Seat",
        {
            "Sessions": dbus.Array([], signature="(so)"),
            "CanGraphical": False,
            "CanMultiSession": True,
            "CanTTY": False,
            "IdleHint": False,
            "ActiveSession": ("", dbus.ObjectPath("/")),
            "Id": seat,
            "IdleSinceHint": dbus.UInt64(0),
            "IdleSinceHintMonotonic": dbus.UInt64(0),
        },
        [("ActivateSession", "s", "", ""), ("Terminate", "", "", "")],
    )

    return seat_path


@dbus.service.method(MOCK_IFACE, in_signature="usb", out_signature="s")
def AddUser(self, uid, username, active):
    """Convenience method to add a user.

    Return the object path of the new user.
    """
    user_path = f"/org/freedesktop/login1/user/{uid}"
    if user_path in mockobject.objects:
        raise dbus.exceptions.DBusException(f"User {uid} already exists", name=MOCK_IFACE + ".UserExists")

    self.AddObject(
        user_path,
        "org.freedesktop.login1.User",
        {
            "DefaultControlGroup": "systemd:/user/" + username,
            "Display": ("", dbus.ObjectPath("/")),
            "GID": dbus.UInt32(uid),
            "IdleHint": False,
            "IdleSinceHint": dbus.UInt64(0),
            "IdleSinceHintMonotonic": dbus.UInt64(0),
            "Linger": False,
            "Name": username,
            "RuntimePath": f"/run/user/{uid}",
            "Service": "",
            "Sessions": dbus.Array([], signature="(so)"),
            "Slice": f"user-{uid}.slice",
            "State": "active" if active else "online",
            "Timestamp": dbus.UInt64(42),
            "TimestampMonotonic": dbus.UInt64(42),
            "UID": dbus.UInt32(uid),
        },
        [
            ("Kill", "s", "", ""),
            ("Terminate", "", "", ""),
        ],
    )

    return user_path


@dbus.service.method(MOCK_IFACE, in_signature="ssusb", out_signature="s")
def AddSession(self, session_id, seat, uid, username, active):
    """Convenience method to add a session.

    If the given seat and/or user do not exit, they will be created.

    Return the object path of the new session.
    """
    seat_path = dbus.ObjectPath(f"/org/freedesktop/login1/seat/{seat}")
    if seat_path not in mockobject.objects:
        self.AddSeat(seat)

    user_path = dbus.ObjectPath(f"/org/freedesktop/login1/user/{uid}")
    if user_path not in mockobject.objects:
        self.AddUser(uid, username, active)

    session_path = dbus.ObjectPath(f"/org/freedesktop/login1/session/{session_id}")
    if session_path in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Session {session_id} already exists", name=MOCK_IFACE + ".SessionExists"
        )

    self.AddObject(
        session_path,
        "org.freedesktop.login1.Session",
        {
            "Controllers": dbus.Array([], signature="s"),
            "ResetControllers": dbus.Array([], signature="s"),
            "Active": active,
            "IdleHint": False,
            "LockedHint": False,
            "KillProcesses": False,
            "Remote": False,
            "Class": "user",
            "DefaultControlGroup": f"systemd:/user/{username}/{session_id}",
            "Display": os.getenv("DISPLAY", ""),
            "Id": session_id,
            "Name": username,
            "RemoteHost": "",
            "RemoteUser": "",
            "Service": "dbusmock",
            "State": "active" if active else "online",
            "TTY": "",
            "Type": "test",
            "Seat": (seat, seat_path),
            "User": (dbus.UInt32(uid), user_path),
            "Audit": dbus.UInt32(0),
            "Leader": dbus.UInt32(1),
            "VTNr": dbus.UInt32(1),
            "IdleSinceHint": dbus.UInt64(0),
            "IdleSinceHintMonotonic": dbus.UInt64(0),
            "Timestamp": dbus.UInt64(42),
            "TimestampMonotonic": dbus.UInt64(42),
        },
        [
            ("Activate", "", "", ""),
            ("Kill", "ss", "", ""),
            ("Lock", "", "", 'self.EmitSignal("", "Lock", "", [])'),
            ("SetIdleHint", "b", "", ""),
            ("SetLockedHint", "b", "", 'self.UpdateProperties("", {"LockedHint": args[0]})'),
            ("Terminate", "", "", ""),
            ("Unlock", "", "", 'self.EmitSignal("", "Unlock", "", [])'),
        ],
    )

    # add session to seat
    obj_seat = mockobject.objects[seat_path]
    cur_sessions = obj_seat.Get("org.freedesktop.login1.Seat", "Sessions")
    cur_sessions.append((session_id, session_path))
    obj_seat.Set("org.freedesktop.login1.Seat", "Sessions", cur_sessions)
    obj_seat.Set("org.freedesktop.login1.Seat", "ActiveSession", (session_id, session_path))

    # add session to user
    obj_user = mockobject.objects[user_path]
    cur_sessions = obj_user.Get("org.freedesktop.login1.User", "Sessions")
    cur_sessions.append((session_id, session_path))
    obj_user.Set("org.freedesktop.login1.User", "Sessions", cur_sessions)

    return session_path



# coding: UTF-8
"""Mock D-Bus objects for test suites."""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import copy
import functools
import importlib
import importlib.util
import os
import sys
import time
import types
from pathlib import Path
from typing import Any, Dict, KeysView, List, Optional, Sequence, Tuple
from xml.etree import ElementTree

import dbus
import dbus.service

# we do not use this ourselves, but mock methods often want to use this
os  # pyflakes pylint: disable=pointless-statement  # noqa: B018

# global path -> DBusMockObject mapping
objects: Dict[str, "DBusMockObject"] = {}

MOCK_IFACE = "org.freedesktop.DBus.Mock"
OBJECT_MANAGER_IFACE = "org.freedesktop.DBus.ObjectManager"


PropsType = Dict[str, Any]
# (in_signature, out_signature, code, dbus_wrapper_fn)
MethodType = Tuple[str, str, str, str]
# (timestamp, method_name, call_args)
CallLogType = Tuple[int, str, Sequence[Any]]


def load_module_from_path(path: Path, template_name: str):
    """Load a mock template from a file path"""
    spec = importlib.util.spec_from_file_location(template_name, path)
    assert spec
    mod = importlib.util.module_from_spec(spec)
    exec(path.read_text("UTF-8"), mod.__dict__, mod.__dict__)  # pylint: disable=exec-used
    return mod


def load_module(name: str):
    """Load a mock template Python module

    This can be a path to the template's .py file, a bare module name in
    ``$XDG_DATA_DIRS/python-dbusmock/templates/``, or a bare module name in dbusmock's shipped templates.
    """
    # specified by path
    pname = Path(name)
    if pname.exists() and pname.suffix == ".py":
        return load_module_from_path(pname, pname.stem)

    # XDG_DATA_DIRS
    xdg_data_dirs = os.environ.get("XDG_DATA_DIRS") or "/usr/local/share/:/usr/share/"
    for d in xdg_data_dirs.split(":"):
        src = Path(d, "python-dbusmock", "templates", name + ".py")
        if src.exists():
            return load_module_from_path(src, name)

    # shipped inside dbusmock package
    return importlib.import_module("dbusmock.templates." + name)


def _format_args(args):
    """Format a D-Bus argument tuple into an appropriate logging string"""

    def format_arg(a):
        if isinstance(a, dbus.Boolean):
            return str(bool(a))
        if isinstance(a, (dbus.Byte, int)):
            return str(int(a))
        if isinstance(a, str):
            return '"' + str(a) + '"'
        if isinstance(a, list):
            return "[" + ", ".join([format_arg(x) for x in a]) + "]"
        if isinstance(a, dict):
            fmta = "{"
            first = True
            for k, v in a.items():
                if first:
                    first = False
                else:
                    fmta += ", "
                fmta += format_arg(k) + ": " + format_arg(v)
            return fmta + "}"

        # fallback
        return repr(a)

    s = ""
    for a in args:
        if s:
            s += " "
        s += format_arg(a)
    if s:
        s = " " + s
    return s


def _wrap_in_dbus_variant(value):
    dbus_types = [
        dbus.types.ByteArray,
        dbus.types.Int16,
        dbus.types.ObjectPath,
        dbus.types.Struct,
        dbus.types.UInt64,
        dbus.types.Boolean,
        dbus.types.Dictionary,
        dbus.types.Int32,
        dbus.types.Signature,
        dbus.types.UInt16,
        dbus.types.UnixFd,
        dbus.types.Byte,
        dbus.types.Double,
        dbus.types.Int64,
        dbus.types.String,
        dbus.types.UInt32,
    ]
    if isinstance(value, dbus.String):
        return dbus.String(str(value), variant_level=1)
    if isinstance(value, dbus.types.Array):
        return value
    if type(value) in dbus_types:
        return type(value)(value.conjugate(), variant_level=1)
    if isinstance(value, str):
        return dbus.String(value, variant_level=1)
    raise dbus.exceptions.DBusException(f"could not wrap type {type(value)}")


def _convert_args(signature: str, args: Tuple[Any, ...]) -> List[Any]:
    """
    Convert types of arguments according to signature, using
    MethodCallMessage.append(); this will also provide type/length
    checks, except for the case of an empty signature
    """
    try:
        if signature == "" and len(args) > 0:
            raise TypeError("Fewer items found in D-Bus signature than in Python arguments")
        m = dbus.connection.MethodCallMessage("a.b", "/a", "a.b", "a")
        m.append(*args, signature=signature)
        return m.get_args_list()
    except Exception as e:
        raise dbus.exceptions.DBusException(
            f"Invalid arguments: {e!s}", name="org.freedesktop.DBus.Error.InvalidArgs"
        ) from e


def loggedmethod(self, func):
    """Decorator for a method to end in the call log"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        fname = func.__name__
        self_arg, args = args[0], args[1:]

        in_signature = getattr(func, "_dbus_in_signature", "")
        args = _convert_args(in_signature, args)

        self.log(fname + _format_args(args))
        self.call_log.append((int(time.time()), fname, args))
        self.MethodCalled(fname, args)

        return func(*[self_arg, *args], **kwargs)

    return wrapper


class DBusMockObject(dbus.service.Object):  # pylint: disable=too-many-instance-attributes
    """Mock D-Bus object

    This can be configured to have arbitrary methods (including code execution)
    and properties via methods on the org.freedesktop.DBus.Mock interface, so
    that you can control the mock from any programming language.

    Beyond that "remote control" API, this is a standard
    `dbus-python service object <https://dbus.freedesktop.org/doc/dbus-python/tutorial.html#exporting-objects>`_.
    """

    def __init__(
        self,
        bus_name: str,
        path: str,
        interface: str,
        props: PropsType,
        logfile: Optional[str] = None,
        is_object_manager: bool = False,
    ) -> None:
        """Create a new DBusMockObject

        :param bus_name: A dbus.service.BusName instance where the object will be put on
        :param path: D-Bus object path
        :param interface: Primary D-Bus interface name of this object (where
                          properties and methods will be put on)
        :param props: A property_name (string) → property (Variant) map with initial
                      properties on "interface"
        :param logfile: When given, method calls will be logged into that file name;
                        if None, logging will be written to stdout. Note that you can
                        also query the called methods over D-Bus with GetCalls() and
                        GetMethodCalls().
        :param is_object_manager: If True, the GetManagedObjects method will
                        automatically be implemented on the object, returning
                        all objects which have this one's path as a prefix of
                        theirs. Note that the InterfacesAdded and
                        InterfacesRemoved signals will not be automatically
                        emitted.
        """
        dbus.service.Object.__init__(self, bus_name, path)

        self.bus_name = bus_name
        self.path = path
        self.interface = interface
        self.is_object_manager = is_object_manager
        self.object_manager: Optional[DBusMockObject] = None

        self._template: Optional[str] = None
        self._template_parameters: Optional[PropsType] = None

        # pylint: disable=consider-using-with
        self.logfile = open(logfile, "wb") if logfile else None  # noqa: SIM115
        self.is_logfile_owner = True
        self.call_log: List[CallLogType] = []

        if props is None:
            props = {}

        self._reset(props)

    def __del__(self) -> None:
        try:
            if self.logfile and self.is_logfile_owner:
                self.logfile.close()
        except AttributeError:
            pass

    def _set_up_object_manager(self) -> None:
        """Set up this mock object as a D-Bus ObjectManager."""
        cond = "k != '/'" if self.path == "/" else f"k.startswith('{self.path}/')"

        code = f"ret = {{dbus.ObjectPath(k): objects[k].props for k in objects.keys() if {cond} }}"
        self.AddMethod(OBJECT_MANAGER_IFACE, "GetManagedObjects", "", "a{oa{sa{sv}}}", code)
        self.object_manager = self

    def _reset(self, props: PropsType) -> None:
        # interface -> name -> value
        self.props = {self.interface: props}

        # interface -> name -> (in_signature, out_signature, code, dbus_wrapper_fn)
        self.methods: Dict[str, Dict[str, MethodType]] = {self.interface: {}}

        if self.is_object_manager:
            self._set_up_object_manager()

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature="ss", out_signature="v")
    def Get(self, interface_name: str, property_name: str) -> Any:
        """Standard D-Bus API for getting a property value"""

        self.log(f"Get {self.path} {interface_name}.{property_name}")

        if not interface_name:
            interface_name = self.interface
        try:
            return self.GetAll(interface_name)[property_name]
        except KeyError as e:
            raise dbus.exceptions.DBusException(
                "no such property " + property_name, name=self.interface + ".UnknownProperty"
            ) from e

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature="s", out_signature="a{sv}")
    def GetAll(self, interface_name: str, *_, **__) -> PropsType:
        """Standard D-Bus API for getting all property values"""

        self.log(f"GetAll {self.path} {interface_name}")

        if not interface_name:
            interface_name = self.interface
        try:
            return self.props[interface_name]
        except KeyError as e:
            raise dbus.exceptions.DBusException(
                "no such interface " + interface_name, name=self.interface + ".UnknownInterface"
            ) from e

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature="ssv", out_signature="")
    def Set(self, interface_name: str, property_name: str, value: Any, *_, **__) -> None:
        """Standard D-Bus API for setting a property value"""

        self.log(f"Set {self.path} {interface_name}.{property_name}{_format_args((value,))}")

        try:
            iface_props = self.props[interface_name]
        except KeyError as e:
            raise dbus.exceptions.DBusException(
                "no such interface " + interface_name, name=self.interface + ".UnknownInterface"
            ) from e

        if property_name not in iface_props:
            raise dbus.exceptions.DBusException(
                "no such property " + property_name, name=self.interface + ".UnknownProperty"
            )

        iface_props[property_name] = value

        self.EmitSignal(
            "org.freedesktop.DBus.Properties",
            "PropertiesChanged",
            "sa{sv}as",
            [interface_name, dbus.Dictionary({property_name: value}, signature="sv"), dbus.Array([], signature="s")],
        )

    @dbus.service.method(MOCK_IFACE, in_signature="ssa{sv}a(ssss)", out_signature="")
    def AddObject(self, path: str, interface: str, properties: PropsType, methods: List[MethodType]) -> None:
        """Dynamically add a new D-Bus object to the mock

        :param path: D-Bus object path
        :param interface: Primary D-Bus interface name of this object (where
                   properties and methods will be put on)
        :param properties: A property_name (string) → value map with initial
                    properties on "interface"
        :param methods: An array of 4-tuples (name, in_sig, out_sig, code) describing
                 methods to add to "interface"; see AddMethod() for details of
                 the tuple values

        If this is a D-Bus ObjectManager instance, the InterfacesAdded signal
        will *not* be emitted for the object automatically; it must be emitted
        manually if desired. This is because AddInterface may be called after
        AddObject, but before the InterfacesAdded signal should be emitted.

        Example::

            dbus_proxy.AddObject('/com/example/Foo/Manager',
                                 'com.example.Foo.Control',
                                 {
                                     'state': dbus.String('online', variant_level=1),
                                 },
                                 [
                                     ('Start', '', '', ''),
                                     ('EchoInt', 'i', 'i', 'ret = args[0]'),
                                     ('GetClients', '', 'ao', 'ret = ["/com/example/Foo/Client1"]'),
                                 ])
        """
        if path in objects:
            raise dbus.exceptions.DBusException(
                f"object {path} already exists", name="org.freedesktop.DBus.Mock.NameError"
            )

        obj = DBusMockObject(self.bus_name, path, interface, properties)
        # make sure created objects inherit the log file stream
        obj.logfile = self.logfile
        obj.object_manager = self.object_manager
        obj.is_logfile_owner = False
        obj.AddMethods(interface, methods)

        objects[path] = obj

    @dbus.service.method(MOCK_IFACE, in_signature="s", out_signature="")
    def RemoveObject(self, path: str) -> None:
        """Remove a D-Bus object from the mock

        As with AddObject, this will *not* emit the InterfacesRemoved signal if
        it's an ObjectManager instance.
        """
        try:
            objects[path].remove_from_connection()
            del objects[path]
        except KeyError as e:
            raise dbus.exceptions.DBusException(
                f"object {path} does not exist", name="org.freedesktop.DBus.Mock.NameError"
            ) from e

    @dbus.service.method(MOCK_IFACE, in_signature="", out_signature="")
    def Reset(self) -> None:
        """Reset the mock object state.

        Remove all mock objects from the bus and tidy up so the state is as if
        python-dbusmock had just been restarted. If the mock object was
        originally created with a template (from the command line, the Python
        API or by calling AddTemplate over D-Bus), it will be
        re-instantiated with that template.
        """
        # Clear other existing objects.
        for obj_name, obj in objects.items():
            if obj_name != self.path:
                obj.remove_from_connection()
        objects.clear()

        # Reinitialise our state. Carefully remove new methods from our dict;
        # they don't not actually exist if they are a statically defined
        # template function
        for method_name in self.methods[self.interface]:
            try:
                delattr(self.__class__, method_name)
            except AttributeError:
                pass

        self._reset({})

        if self._template is not None:
            self.AddTemplate(self._template, self._template_parameters)

        objects[self.path] = self

    @dbus.service.method(MOCK_IFACE, in_signature="sssss", out_signature="")
    def AddMethod(self, interface, name: str, in_sig: str, out_sig: str, code: str) -> None:
        """Dynamically add a method to this object

        :param interface: D-Bus interface to add this to. For convenience you can
                          specify '' here to add the method to the object's main
                          interface (as specified on construction).
        :param name: Name of the method
        :param in_sig: Signature of input arguments; for example "ias" for a method
                       that takes an int32 and a string array as arguments; see
                       `the DBus spec <http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-signatures>`_.
        :param out_sig: Signature of output arguments; for example "s" for a method
                       that returns a string; use '' for methods that do not return
                       anything.
        :param code: Python 3 code to run in the method call; you have access to the
                     arguments through the "args" list, and can set the return value
                     by assigning a value to the "ret" variable. You can also read the
                     global "objects" variable, which is a dictionary mapping object
                     paths to DBusMockObject instances.

                     For keeping state across method calls, you are free to use normal
                     Python members of the "self" object, which will be persistent for
                     the whole mock's life time. E. g. you can have a method with
                     "self.my_state = True", and another method that returns it with
                     "ret = self.my_state".

                     Methods can raise exceptions in the usual way, in particular
                     `dbus.exceptions.DBusException <https://dbus.freedesktop.org/doc/dbus-python/dbus.html#dbus.DBusException>`_.

                     When specifying '', the method will not do anything (except
                     logging) and return None.


        This is meant for adding a method to a mock at runtime, from any programming language.
        You can also use it in templates in the load() function.

        For implementing non-trivial and static methods in templates, it is recommended to
        implement them in the normal dbus-python way with using the @dbus.service.method
        decorator instead.
        """
        # pylint: disable=protected-access

        if not interface:
            interface = self.interface
        n_args = len(dbus.Signature(in_sig))

        # we need to have separate methods for dbus-python, so clone
        # mock_method(); using message_keyword with this dynamic approach fails
        # because inspect cannot handle those, so pass on interface and method
        # name as first positional arguments
        # pylint: disable=unnecessary-lambda-assignment
        method = lambda self, *args, **kwargs: DBusMockObject.mock_method(  # noqa: E731
            self, interface, name, in_sig, *args, **kwargs
        )

        # we cannot specify in_signature here, as that trips over a consistency
        # check in dbus-python; we need to set it manually instead
        dbus_method = dbus.service.method(interface, out_signature=out_sig)(method)
        dbus_method.__name__ = str(name)
        dbus_method._dbus_in_signature = in_sig
        dbus_method._dbus_args = [f"arg{i}" for i in range(1, n_args + 1)]

        # for convenience, add mocked methods on the primary interface as
        # callable methods
        if interface == self.interface:
            setattr(self.__class__, name, dbus_method)

        self.methods.setdefault(interface, {})[str(name)] = (in_sig, out_sig, code, dbus_method)

    @dbus.service.method(MOCK_IFACE, in_signature="sa(ssss)", out_signature="")
    def AddMethods(self, interface: str, methods: List[MethodType]) -> None:
        """Add several methods to this object

        :param interface: D-Bus interface to add this to. For convenience you can
                          specify '' here to add the method to the object's main
                          interface (as specified on construction).
        :param methods: list of 4-tuples (name, in_sig, out_sig, code) describing one
                        method each. See AddMethod() for details of the tuple values.
        """
        for method in methods:
            self.AddMethod(interface, *method)

    def _set_property(self, interface, name, value):
        # copy.copy removes one level of variant-ness, which means that the
        # types get exported in introspection data correctly, but we can't do
        # this for container types.
        if not isinstance(value, (dbus.Dictionary, dbus.Array)):
            value = copy.copy(value)

        self.props.setdefault(interface, {})[name] = value

    @dbus.service.method(MOCK_IFACE, in_signature="sa{sv}", out_signature="")
    def UpdateProperties(self, interface: str, properties: PropsType) -> None:
        """Update properties on this object and send a PropertiesChanged signal

        :param interface: D-Bus interface to update this to. For convenience you can
                          specify '' here to add the property to the object's main
                          interface (as specified on construction).
        :param properties: A property_name (string) → value map
        """
        changed_props = {}

        for name, value in properties.items():
            if not interface:
                interface = self.interface
            if name not in self.props.get(interface, {}):
                raise dbus.exceptions.DBusException(f"property {name} not found", name=interface + ".NoSuchProperty")

            self._set_property(interface, name, value)
            changed_props[name] = _wrap_in_dbus_variant(value)

        self.EmitSignal(dbus.PROPERTIES_IFACE, "PropertiesChanged", "sa{sv}as", [interface, changed_props, []])

    @dbus.service.method(MOCK_IFACE, in_signature="ssv", out_signature="")
    def AddProperty(self, interface: str, name: str, value: Any) -> None:
        """Add property to this object

        :param interface: D-Bus interface to add this to. For convenience you can
                           specify '' here to add the property to the object's main
                           interface (as specified on construction).
        :param name: Property name.
        :param value: Property value.
        """
        if not interface:
            interface = self.interface
        if name in self.props.get(interface, {}):
            raise dbus.exceptions.DBusException(
                f"property {name} already exists", name=self.interface + ".PropertyExists"
            )

        self._set_property(interface, name, value)

    @dbus.service.method(MOCK_IFACE, in_signature="sa{sv}", out_signature="")
    def AddProperties(self, interface: str, properties: PropsType) -> None:
        """Add several properties to this object

        :param interface: D-Bus interface to add this to. For convenience you can
                          specify '' here to add the property to the object's main
                          interface (as specified on construction).
        :param properties: A property_name (string) → value map
        """
        for k, v in properties.items():
            self.AddProperty(interface, k, v)

    @dbus.service.method(MOCK_IFACE, in_signature="sa{sv}", out_signature="")
    def AddTemplate(self, template: str, parameters: PropsType) -> None:
        """Load a template into the mock.

        python-dbusmock ships a set of standard mocks for common system
        services such as UPower and NetworkManager. With these the actual tests
        become a lot simpler, as they only have to set up the particular
        properties for the tests, and not the skeleton of common properties,
        interfaces, and methods.

        :param template: Name of the template to load or the full path to a *.py file
                         for custom templates. See "pydoc dbusmock.templates" for a
                         list of available templates from python-dbusmock package, and
                         "pydoc dbusmock.templates.NAME" for documentation about
                         template NAME.
        :param parameters: A parameter (string) → value (variant) map, for
                           parameterizing templates. Each template can define their
                           own, see documentation of that particular template for
                           details.
        """
        try:
            module = load_module(template)
        except ImportError as e:
            raise dbus.exceptions.DBusException(
                f"Cannot add template {template}: {e!s}", name="org.freedesktop.DBus.Mock.TemplateError"
            ) from e

        # If the template specifies this is an ObjectManager, set that up
        if hasattr(module, "IS_OBJECT_MANAGER") and module.IS_OBJECT_MANAGER:
            self._set_up_object_manager()

        # pick out all D-Bus service methods and add them to our interface
        for symbol in dir(module):
            # pylint: disable=protected-access
            fn = getattr(module, symbol)
            if "_dbus_interface" in dir(fn) and ("_dbus_is_signal" not in dir(fn) or not fn._dbus_is_signal):
                fn = loggedmethod(self, fn)

                # for dbus-python compatibility, add methods as callables
                setattr(self.__class__, symbol, fn)
                self.methods.setdefault(fn._dbus_interface, {})[str(symbol)] = (
                    fn._dbus_in_signature,
                    fn._dbus_out_signature,
                    "",
                    fn,
                )

        if parameters is None:
            parameters = {}

        module.load(self, parameters)
        # save the given template and parameters for re-instantiation on
        # Reset()
        self._template = template
        self._template_parameters = parameters

    def _emit_signal(
        self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...], details: PropsType
    ) -> None:
        # pylint: disable=protected-access
        if not interface:
            interface = self.interface

        args = _convert_args(signature, sigargs)

        path = details.get("path", self.path)
        sig = dbus.lowlevel.SignalMessage(path, interface, name)
        sig.append(*args, signature=signature)
        dest = details.get("destination", None)
        if dest is not None:
            sig.set_destination(dest)

        for location in self.locations:
            conn = location[0]
            conn.send_message(sig)
        self.log(f"emit {path} {interface}.{name}{_format_args(args)}")

    @dbus.service.method(MOCK_IFACE, in_signature="sssav", out_signature="")
    def EmitSignal(self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...]) -> None:
        """Emit a signal from the object.

        :param interface: D-Bus interface to send the signal from. For convenience you
                          can specify '' here to add the method to the object's main
                          interface (as specified on construction).
        :param name: Name of the signal
        :param signature: Signature of input arguments; for example "ias" for a signal
                          that takes an int32 and a string array as arguments; see
                          http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-signatures
        :param args: variant array with signal arguments; must match order and type in
                     "signature"
        """
        self._emit_signal(interface, name, signature, sigargs, {})

    @dbus.service.method(MOCK_IFACE, in_signature="sssava{sv}", out_signature="")
    def EmitSignalDetailed(
        self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...], details: PropsType
    ) -> None:
        """Emit a signal from the object with extra details.

        :param interface: D-Bus interface to send the signal from. For convenience you
                          can specify '' here to add the method to the object's main
                          interface (as specified on construction).
        :param name: Name of the signal
        :param signature: Signature of input arguments; for example "ias" for a signal
                       that takes an int32 and a string array as arguments; see
                       http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-signatures
        :param args: variant array with signal arguments; must match order and type in
                     "signature"
        :param details: dictionary with a string key/value entries. Supported keys are:
                   "destination": for the signal destination
                   "path": for the object path to send the signal from
        """
        self._emit_signal(interface, name, signature, sigargs, details)

    @dbus.service.method(MOCK_IFACE, in_signature="", out_signature="a(tsav)")
    def GetCalls(self) -> List[CallLogType]:
        """List all the logged calls since the last call to ClearCalls().

        Return a list of (timestamp, method_name, args_list) tuples.
        """
        return self.call_log

    @dbus.service.method(MOCK_IFACE, in_signature="s", out_signature="a(tav)")
    def GetMethodCalls(self, method: str) -> List[Tuple[int, Sequence[Any]]]:
        """List all the logged calls of a particular method.

        Return a list of (timestamp, args_list) tuples.
        """
        return [(row[0], row[2]) for row in self.call_log if row[1] == method]

    @dbus.service.method(MOCK_IFACE, in_signature="", out_signature="")
    def ClearCalls(self) -> None:
        """Empty the log of mock call signatures."""

        self.call_log = []

    @dbus.service.signal(MOCK_IFACE, signature="sav")
    def MethodCalled(self, name, args):
        """Signal emitted for every called mock method.

        This is emitted for all mock method calls.  This can be used to confirm
        that a particular method was called with particular arguments, as an
        alternative to reading the mock's log or GetCalls().
        """

    def object_manager_emit_added(self, path: str) -> None:
        """Emit ObjectManager.InterfacesAdded signal"""

        if self.object_manager is not None:
            self.object_manager.EmitSignal(
                OBJECT_MANAGER_IFACE, "InterfacesAdded", "oa{sa{sv}}", [dbus.ObjectPath(path), objects[path].props]
            )

    def object_manager_emit_removed(self, path: str) -> None:
        """Emit ObjectManager.InterfacesRemoved signal"""

        if self.object_manager is not None:
            self.object_manager.EmitSignal(
                OBJECT_MANAGER_IFACE, "InterfacesRemoved", "oas", [dbus.ObjectPath(path), objects[path].props]
            )

    def mock_method(self, interface: str, dbus_method: str, in_signature: str, *m_args, **_) -> Any:
        """Master mock method.

        This gets "instantiated" in AddMethod(). Execute the code snippet of
        the method and return the "ret" variable if it was set.
        """
        # print('mock_method', dbus_method, self, in_signature, args, _, file=sys.stderr)

        try:
            args = _convert_args(in_signature, m_args)

            self.log(dbus_method + _format_args(args))
            self.call_log.append((int(time.time()), str(dbus_method), args))
            self.MethodCalled(dbus_method, args)

            # The code may be a Python 3 string to interpret, or may be a function
            # object (if AddMethod was called from within Python itself, rather than
            # over D-Bus).
            code = self.methods[interface][dbus_method][2]
            if code and isinstance(code, types.FunctionType):
                return code(self, *args)
            if code:
                loc = locals().copy()
                exec(code, globals(), loc)  # pylint: disable=exec-used
                if "ret" in loc:
                    return loc["ret"]
        except Exception as e:
            self.log(dbus_method + " raised: " + str(e))
            raise e

        return None

    def log(self, msg: str) -> None:
        """Log a message, prefixed with a timestamp.

        If a log file was specified in the constructor, it is written there,
        otherwise it goes to stdout.
        """
        fd = self.logfile.fileno() if self.logfile else sys.stdout.fileno()

        os.write(fd, f"{time.time():.3f} {msg}\n".encode())

    @dbus.service.method(
        dbus.INTROSPECTABLE_IFACE,
        in_signature="",
        out_signature="s",
        path_keyword="object_path",
        connection_keyword="connection",
    )
    def Introspect(self, object_path: str, connection: dbus.connection.Connection) -> str:
        """Return XML description of this object's interfaces, methods and signals.

        This wraps dbus-python's Introspect() method to include the dynamic
        methods and properties.
        """
        # _dbus_class_table is an indirect private member of dbus.service.Object that pylint fails to see
        # pylint: disable=no-member

        # temporarily add our dynamic methods
        cls = self.__class__.__module__ + "." + self.__class__.__name__
        orig_interfaces = self._dbus_class_table[cls]

        mock_interfaces = orig_interfaces.copy()
        for iface, methods in self.methods.items():
            for method, impl in methods.items():
                mock_interfaces.setdefault(iface, {})[method] = impl[3]
        self._dbus_class_table[cls] = mock_interfaces

        xml = dbus.service.Object.Introspect(self, object_path, connection)

        tree = ElementTree.fromstring(xml)

        for name, name_props in self.props.items():
            # We might have properties for new interfaces we don't know about
            # yet. Try to find an existing <interface> node named after our
            # interface to append to, and create one if we can't.
            interface = tree.find(f".//interface[@name='{name}']")
            if interface is None:
                interface = ElementTree.Element("interface", {"name": name})
                tree.append(interface)

            for prop, val in name_props.items():
                if val is None:
                    # can't guess type from None, skip
                    continue
                elem = ElementTree.Element(
                    "property",
                    {
                        "name": prop,
                        # We don't store the signature anywhere, so guess it.
                        "type": dbus.lowlevel.Message.guess_signature(val),
                        "access": "readwrite",
                    },
                )

                interface.append(elem)

        xml = ElementTree.tostring(tree, encoding="utf8", method="xml").decode("utf8")

        # restore original class table
        self._dbus_class_table[cls] = orig_interfaces

        return xml


# Overwrite dbus-python's _method_lookup(), as that offers no way to have the
# same method name on different interfaces
orig_method_lookup = dbus.service._method_lookup  # pylint: disable=protected-access


def _dbusmock_method_lookup(obj, method_name, dbus_interface):
    try:
        m = obj.methods[dbus_interface or obj.interface][method_name]
        return (m[3], m[3])
    except KeyError:
        return orig_method_lookup(obj, method_name, dbus_interface)


dbus.service._method_lookup = _dbusmock_method_lookup  # pylint: disable=protected-access


#
# Helper API for templates
#


def get_objects() -> KeysView[str]:
    """Return all existing object paths"""

    return objects.keys()


def get_object(path) -> DBusMockObject:
    """Return object for a given object path"""

    return objects[path]



"""unittest.TestCase convenience methods for DBusMocks"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import enum
import errno
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import unittest
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import dbus
import dbus.proxies

from dbusmock.mockobject import MOCK_IFACE, OBJECT_MANAGER_IFACE, load_module


class BusType(enum.Enum):
    """Represents a system or session bus"""

    SESSION = "session"
    SYSTEM = "system"

    @property
    def environ(self) -> Tuple[str, Optional[str]]:
        """Returns the name and value of this bus' address environment variable"""
        env = f"DBUS_{self.value.upper()}_BUS_ADDRESS"
        value = os.environ.get(env)
        return env, value

    def get_connection(self) -> dbus.bus.Connection:
        """Get a dbus.bus.BusConnection() object to this bus.

        This uses the current environment variables for this bus (if any) and falls back
        to dbus.SystemBus() or dbus.SessionBus() otherwise.

        This is preferrable to dbus.SystemBus() and dbus.SessionBus() as those
        do not get along with multiple changing local test buses.
        """
        _, val = self.environ
        if val:
            return dbus.bus.BusConnection(val)
        if self == BusType.SYSTEM:
            return dbus.SystemBus()
        return dbus.SessionBus()

    def reload_configuration(self):
        """Notify this bus that it needs to reload the configuration"""
        bus = self.get_connection()
        dbus_obj = bus.get_object("org.freedesktop.DBus", "/org/freedesktop/DBus")
        dbus_if = dbus.Interface(dbus_obj, "org.freedesktop.DBus")
        dbus_if.ReloadConfig()

    def wait_for_bus_object(self, dest: str, path: str, timeout: float = 60.0):
        """Wait for an object to appear on D-Bus

        Raise an exception if object does not appear within one minute. You can
        change the timeout in seconds with the "timeout" keyword argument.
        """
        bus = self.get_connection()

        last_exc = None
        # we check whether the name is owned first, to avoid race conditions
        # with service activation; once it's owned, wait until we can actually
        # call methods
        while timeout > 0:
            if bus.name_has_owner(dest):
                try:
                    p = dbus.Interface(bus.get_object(dest, path), dbus_interface=dbus.INTROSPECTABLE_IFACE)
                    p.Introspect()
                    break
                except dbus.exceptions.DBusException as e:
                    last_exc = e
                    if ".UnknownInterface" in str(e):
                        break

            timeout -= 0.1
            time.sleep(0.1)
        if timeout <= 0:
            assert timeout > 0, f"timed out waiting for D-Bus object {path}: {last_exc}"


class PrivateDBus:
    """A D-Bus daemon instance that represents a private session or system bus.

    If used as a context manager it will automatically start the bus and clean up
    after itself on exit:

        >>> with PrivateDBus(BusType.SESSION) as bus:
        >>>    do_something(bus)

    Otherwise, `start()` and `stop()` manually.
    """

    def __init__(self, bustype: BusType):
        self.bustype = bustype
        self._daemon: Optional[subprocess.Popen] = None

        self._datadir = Path(tempfile.mkdtemp(prefix="dbusmock_data_"))
        self._socket = self._datadir / f"{self.bustype.value}_bus.socket"
        subdir = "system-services" if bustype == BusType.SYSTEM else "services"
        self._servicedir = self._datadir / subdir
        self._servicedir.mkdir(parents=True)

        self._config = self._servicedir / f"dbusmock_{self.bustype.value}_cfg"
        self._config.write_text(
            f"""<!DOCTYPE busconfig PUBLIC "-//freedesktop//DTD D-Bus Bus Configuration 1.0//EN"
     "http://www.freedesktop.org/standards/dbus/1.0/busconfig.dtd">
    <busconfig>
      <type>{self.bustype.value}</type>
      <keep_umask/>
      <listen>unix:path={self._socket}</listen>
      <!-- We do not add standard_{self.bustype.value}_servicedirs (i.e. we only have our private services directory). -->
      <servicedir>{self._servicedir}</servicedir>
      <policy context="default">
        <allow send_destination="*" eavesdrop="true"/>
        <allow eavesdrop="true"/>
        <allow own="*"/>
      </policy>
    </busconfig>
    """
        )

    def __enter__(self) -> "PrivateDBus":
        # Allow for start() to be called manually even before the `with`
        if self._daemon is None:
            self.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Allow for stop() to be called manually within `with`
        if self._daemon is not None:
            self.stop()

    @property
    def address(self) -> str:
        """Returns this D-Bus' address in the environment variable format, i.e. something like
        unix:path=/path/to/socket
        """
        assert self._daemon is not None, "Call start() first"
        return f"unix:path={self._socket}"

    @property
    def servicedir(self) -> Path:
        """The services directory (full path) for any ``.service`` files that need to be known to
        this D-Bus.
        """
        return self._servicedir

    @property
    def pid(self) -> int:
        """Return the pid of this D-Bus daemon process"""
        assert self._daemon is not None, "Call start() first"
        return self._daemon.pid

    def start(self):
        """Start the D-Bus daemon"""
        argv = ["dbus-daemon", f"--config-file={self._config}"]
        # pylint: disable=consider-using-with
        self._daemon = subprocess.Popen(argv)
        for _ in range(10):
            if self._socket.exists():
                break
            time.sleep(0.1)
        else:
            assert self._socket.exists(), "D-Bus socket never created"

        env, _ = self.bustype.environ
        os.environ[env] = self.address

    def stop(self):
        """Stop the D-Bus daemon"""
        if self._daemon:
            try:
                self._daemon.terminate()
                try:
                    self._daemon.wait(timeout=1)
                except subprocess.TimeoutExpired:
                    self._daemon.kill()
            except ProcessLookupError:
                pass
            self._daemon = None

        shutil.rmtree(self._datadir, ignore_errors=True)

    def enable_service(self, service: str):
        """Enable the given well-known service name inside dbusmock

        This symlinks a service file from the usual dbus service directories
        into the dbusmock environment. Doing that allows the service to be
        launched automatically if they are defined within $XDG_DATA_DIRS.

        The daemon configuration is reloaded if a test bus is running.
        """
        xdg_data_dirs = os.environ.get("XDG_DATA_DIRS") or "/usr/local/share/:/usr/share/"
        subdir = "system-services" if self.bustype == BusType.SYSTEM else "services"
        for d in xdg_data_dirs.split(":"):
            src = Path(d) / "dbus-1" / subdir / f"{service}.service"
            if src.exists():
                assert self._servicedir.exists()
                (self._servicedir / f"{service}.service").symlink_to(src)
                break
        else:
            raise AssertionError(f"Service {service} not found in XDG_DATA_DIRS ({xdg_data_dirs})")

        if self._daemon:
            self.bustype.reload_configuration()

    def disable_service(self, service):
        """Disable the given well known service name inside dbusmock

        This unlink's the .service file for the service and reloads the
        daemon configuration if a test bus is running.
        """
        try:
            (self._servicedir / f"{service}.service").unlink()
        except OSError:
            raise AssertionError(f"Service {service} not found") from None

        if self._daemon:
            self.bustype.reload_configuration()


class DBusTestCase(unittest.TestCase):
    """Base class for D-Bus mock tests.

    This provides some convenience API to start/stop local D-Buses, so that you
    can run a private local session and/or system bus to run mocks on.

    This also provides a spawn_server() static method to run the D-Bus mock
    server in a separate process.
    """

    session_bus_pid = None
    system_bus_pid = None
    _DBusTestCase__datadir = None
    _busses: Dict[BusType, PrivateDBus] = {
        BusType.SESSION: None,  # type: ignore[dict-item]
        BusType.SYSTEM: None,  # type: ignore[dict-item]
    }

    @staticmethod
    def _bus(bustype: BusType) -> PrivateDBus:
        """Return (and create if necessary) the singleton DBus for the given bus type"""
        if not DBusTestCase._busses.get(bustype):
            DBusTestCase._busses[bustype] = PrivateDBus(bustype)
        return DBusTestCase._busses[bustype]

    @staticmethod
    def get_services_dir(system_bus: bool = False) -> str:
        """Returns the private services directory for the bus type in question.
        This allows dropping in a .service file so that the dbus server inside
        dbusmock can launch it.
        """
        bus = DBusTestCase._bus(bustype=BusType.SYSTEM if system_bus else BusType.SESSION)
        return str(bus.servicedir)

    @classmethod
    def tearDownClass(cls):
        for bustype in BusType:
            bus = DBusTestCase._busses.get(bustype)
            if bus:
                bus.stop()
                setattr(DBusTestCase, f"{bustype.value}_bus_pid", None)
                del DBusTestCase._busses[bustype]

    @classmethod
    def __start_bus(cls, bus_type) -> None:
        bustype = BusType(bus_type)
        old_pid = getattr(DBusTestCase, f"{bustype.value}_bus_pid")
        assert old_pid is None, f"PID {old_pid} still alive?"
        assert DBusTestCase._busses.get(bustype) is None
        bus = DBusTestCase._bus(bustype)
        bus.start()
        setattr(DBusTestCase, f"{bustype.value}_bus_pid", bus.pid)

    @classmethod
    def start_session_bus(cls) -> None:
        """Set up a private local session bus

        This gets stopped automatically at class teardown.
        """
        cls.__start_bus("session")

    @classmethod
    def start_system_bus(cls) -> None:
        """Set up a private local system bus

        This gets stopped automatically at class teardown.
        """
        cls.__start_bus("system")

    @staticmethod
    def start_dbus(conf: Optional[str] = None) -> Tuple[int, str]:
        """Start a D-Bus daemon

        Return (pid, address) pair.

        Normally you do not need to call this directly. Use start_system_bus()
        and start_session_bus() instead.
        """
        argv = ["dbus-daemon", "--fork", "--print-address=1", "--print-pid=1"]
        if conf:
            argv.append("--config-file=" + str(conf))
        else:
            argv.append("--session")
        lines = subprocess.check_output(argv, universal_newlines=True).strip().splitlines()
        assert len(lines) == 2, "expected exactly 2 lines of output from dbus-daemon"
        # usually the first line is the address, but be lenient and accept any order
        try:
            return (int(lines[1]), lines[0])
        except ValueError:
            return (int(lines[0]), lines[1])

    @staticmethod
    def stop_dbus(pid: int) -> None:
        """Stop a D-Bus daemon

        Normally you do not need to call this directly. When you use
        start_system_bus() and start_session_bus(), these buses are
        automatically stopped in tearDownClass().
        """
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        for _ in range(50):
            try:
                os.kill(pid, signal.SIGTERM)
                os.waitpid(pid, os.WNOHANG)
            except ChildProcessError:
                break
            except OSError as e:
                if e.errno == errno.ESRCH:
                    break
                raise
            time.sleep(0.1)
        else:
            sys.stderr.write("ERROR: timed out waiting for bus process to terminate\n")
            os.kill(pid, signal.SIGKILL)
            try:
                os.waitpid(pid, 0)
            except ChildProcessError:
                pass
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

    @staticmethod
    def get_dbus(system_bus: bool = False) -> dbus.Bus:
        """Get a dbus.bus.BusConnection() object to this bus

        This is preferrable to dbus.SystemBus() and dbus.SessionBus() as those
        do not get along with multiple changing local test buses.

        This is a legacy method kept for backwards compatibility, use
        BusType.get_connection() instead.
        """
        bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        return bustype.get_connection()

    @staticmethod
    def wait_for_bus_object(dest: str, path: str, system_bus: bool = False, timeout: int = 600):
        """Wait for an object to appear on D-Bus

        Raise an exception if object does not appear within one minute. You can
        change the timeout with the "timeout" keyword argument which specifies
        deciseconds.

        This is a legacy method kept for backwards compatibility, use
        BusType.wait_for_bus_object() instead.
        """
        bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        bustype.wait_for_bus_object(dest, path, timeout / 10.0)

    @staticmethod
    def spawn_server(name: str, path: str, interface: str, system_bus: bool = False, stdout=None) -> subprocess.Popen:
        """Run a DBusMockObject instance in a separate process

        The daemon will terminate automatically when the D-Bus that it connects
        to goes down.  If that does not happen (e. g. you test on the actual
        system/session bus), you need to kill it manually.

        This function blocks until the spawned DBusMockObject is ready and
        listening on the bus.

        Returns the Popen object of the spawned daemon.

        This is a legacy method kept for backwards compatibility,
        use SpawnedMock.spawn_for_name() instead.
        """
        bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        server = SpawnedMock.spawn_for_name(name, path, interface, bustype, stdout=stdout, stderr=None)
        return server.process

    @staticmethod
    def spawn_server_template(
        template: str, parameters: Optional[Dict[str, Any]] = None, stdout=None, system_bus: Optional[bool] = None
    ) -> Tuple[subprocess.Popen, dbus.proxies.ProxyObject]:
        """Run a D-Bus mock template instance in a separate process

        This starts a D-Bus mock process and loads the given template with
        (optional) parameters into it. For details about templates see
        dbusmock.DBusMockObject.AddTemplate().

        Usually a template should specify SYSTEM_BUS = False/True to select whether it
        gets loaded on the session or system bus. This can be overridden with the system_bus
        parameter. For templates which don't set SYSTEM_BUS, this parameter has to be set.

        The daemon will terminate automatically when the D-Bus that it connects
        to goes down.  If that does not happen (e. g. you test on the actual
        system/session bus), you need to kill it manually.

        This function blocks until the spawned DBusMockObject is ready and
        listening on the bus.

        Returns a pair (daemon Popen object, main dbus object).

        This is a legacy method kept for backwards compatibility,
        use SpawnedMock.spawn_with_template() instead.
        """
        if system_bus is not None:  # noqa: SIM108
            bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        else:
            bustype = None
        server = SpawnedMock.spawn_with_template(template, parameters, bustype, stdout, stderr=None)
        return server.process, server.obj

    @staticmethod
    def enable_service(service, system_bus: bool = False) -> None:
        """Enable the given well known service name inside dbusmock

        This symlinks a service file from the usual dbus service directories
        into the dbusmock environment. Doing that allows the service to be
        launched automatically if they are defined within $XDG_DATA_DIRS.

        The daemon configuration is reloaded if a test bus is running.

        This is a legacy method kept for backwards compatibility. Use
        PrivateDBus.enable_service() instead.
        """
        bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        bus = DBusTestCase._bus(bustype)
        bus.enable_service(service)

    @staticmethod
    def disable_service(service, system_bus: bool = False) -> None:
        """Disable the given well known service name inside dbusmock

        This unlink's the .service file for the service and reloads the
        daemon configuration if a test bus is running.
        """
        bustype = BusType.SYSTEM if system_bus else BusType.SESSION
        bus = DBusTestCase._bus(bustype)
        bus.disable_service(service)


class SpawnedMock:
    """
    An instance of a D-Bus mock template instance in a separate process.

    See SpawnedMock.spawn_for_name() and SpawnedMock.spawn_with_template()
    the typical entry points.
    """

    def __init__(self, process: subprocess.Popen, obj: dbus.proxies.ProxyObject):
        self._process = process
        self._process_is_running = True
        self._obj = obj

    @property
    def process(self) -> subprocess.Popen:
        """Returns the process that is this mock template"""
        return self._process

    @property
    def obj(self):
        """The D-Bus object this server was spawned for"""
        return self._obj

    def __enter__(self) -> "SpawnedMock":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    def terminate(self):
        """Terminate the process"""
        if self._process.returncode is None:
            self._process.poll()

        if self._process.returncode is None:
            if self._process.stdout:
                self._process.stdout.close()
            if self._process.stderr:
                self._process.stderr.close()
            try:
                self._process.terminate()
                try:
                    self._process.wait(timeout=1)
                except subprocess.TimeoutExpired:
                    self._process.kill()
            except ProcessLookupError:
                pass

    @property
    def stdout(self):
        """
        The stdout of the process, if no caller-specific stdout
        was specified in spawn_for_name() or spawn_with_template().
        """
        return self._process.stdout

    @property
    def stderr(self):
        """
        The stderr of the process, if no caller-specific stderr
        was specified in spawn_for_name() or spawn_with_template().
        """
        return self._process.stderr

    @classmethod
    def spawn_for_name(
        cls,
        name: str,
        path: str,
        interface: str,
        bustype: BusType = BusType.SESSION,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) -> "SpawnedMock":
        """Run a DBusMockObject instance in a separate process

        The daemon will terminate automatically when the D-Bus that it connects
        to goes down.  If that does not happen (e. g. you test on the actual
        system/session bus), you need to kill it manually.

        This function blocks until the spawned DBusMockObject is ready and
        listening on the bus.

        By default, stdout and stderr of the spawned process is available via the
        SpawnedMock.stdout and SpawnedMock.stderr properties on the returned object.
        """
        argv = [sys.executable, "-m", "dbusmock", f"--{bustype.value}", name, path, interface]
        bus = bustype.get_connection()
        if bus.name_has_owner(name):
            raise AssertionError(f"Trying to spawn a server for name {name} but it is already owned!")

        # pylint: disable=consider-using-with
        daemon = subprocess.Popen(argv, stdout=stdout, stderr=stderr)

        # wait for daemon to start up
        bustype.wait_for_bus_object(name, path)
        obj = bus.get_object(name, path)

        return cls(process=daemon, obj=obj)

    @classmethod
    def spawn_with_template(
        cls,
        template: str,
        parameters: Optional[Dict[str, Any]] = None,
        bustype: Optional[BusType] = None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ):
        """Run a D-Bus mock template instance in a separate process

        This starts a D-Bus mock process and loads the given template with
        (optional) parameters into it. For details about templates see
        dbusmock.DBusMockObject.AddTemplate().

        Usually a template should specify SYSTEM_BUS = False/True to select whether it
        gets loaded on the session or system bus. This can be overridden with the system_bus
        parameter. For templates which don't set SYSTEM_BUS, this parameter has to be set.

        The daemon will terminate automatically when the D-Bus that it connects
        to goes down.  If that does not happen (e. g. you test on the actual
        system/session bus), you need to kill it manually.

        This function blocks until the spawned DBusMockObject is ready and
        listening on the bus.

        Returns a pair (daemon Popen object, main dbus object).
        """

        # we need the bus address from the template module
        module = load_module(template)

        is_object_manager = module.IS_OBJECT_MANAGER if hasattr(module, "IS_OBJECT_MANAGER") else False

        if is_object_manager and not hasattr(module, "MAIN_IFACE"):  # noqa: SIM108
            interface_name = OBJECT_MANAGER_IFACE
        else:
            interface_name = module.MAIN_IFACE

        if bustype is None:
            bustype = BusType.SYSTEM if module.SYSTEM_BUS else BusType.SESSION

        assert bustype is not None

        server = SpawnedMock.spawn_for_name(module.BUS_NAME, module.MAIN_OBJ, interface_name, bustype, stdout, stderr)
        if not parameters:
            parameters = dbus.Dictionary({}, signature="sv")
        server.obj.AddTemplate(template, parameters, dbus_interface=MOCK_IFACE)
        return server
